from rjsmin import jsmin

from menagerie.Items.MinifiedItemMixin import MinifiedItemMixin
from menagerie.Items.Static.StaticItem import StaticItem


class StaticJS(MinifiedItemMixin, StaticItem):
    extensions = ('js',)
    minify_key = 'js'
    out_extension = 'min.js'
    byte_mode = False

    def minify(self, content: str) -> str:
        return jsmin(content)



import hashlib
from json import loads, dumps
from pathlib import Path
from shutil import rmtree

import menagerie
from menagerie.Items.AbstractItem import AbstractItem

__all__ = ('CacheHandler',)

from menagerie.Settings import Settings


class CacheHandler:
    HASH_ALGORITHM = 'sha256'

    def __init__(self, folder: Path, settings: Settings):
        self.settings = settings
        if self.settings['cache_enabled'] is False:
            return

        self.cache_folder = folder
        self.hash_file = Path(self.cache_folder, "hashes.json")
        self.out_cache = Path(self.cache_folder, "out_cache/")
        self.cache_folder.mkdir(parents=True, exist_ok=True)
        self.out_cache.mkdir(parents=True, exist_ok=True)

        if self.hash_file.exists():
            self.hashes = loads(self.hash_file.read_text(encoding='utf-8'))
        else:
            self.hashes = {'version': menagerie.__version__, 'config': "", 'meta': None, 'items': {}}

        config_hash = self.get_hash(settings['config_path'].read_bytes())
        if self.hashes['config'] != config_hash:
            self.invalidate()
            self.hashes['config'] = config_hash
        if self.hashes['version'] != menagerie.__version__:
            self.invalidate()

    def invalidate(self):
        rmtree(self.out_cache)
        self.out_cache.mkdir(parents=True, exist_ok=True)
        self.hashes = {'version': menagerie.__version__, 'config': "", 'meta': None, 'items': {}}

    def get_hash(self, content: bytes) -> str:
        hasher = hashlib.new(self.HASH_ALGORITHM)
        hasher.update(content)
        return hasher.hexdigest()

    def get_out_cache_path(self, out_path: Path):
        return Path(self.out_cache, out_path.relative_to(self.settings['out_dir']))

    def check_for_hit(self, rel_path: str, content: str) -> [bool, str]:
        hashed_content = self.get_hash(bytes(content, 'utf-8'))
        return (rel_path in self.hashes['items'].keys() and self.hashes['items'][rel_path] == hashed_content), hashed_content

    def get_cached_output(self, out_path: Path) -> str:
        return self.get_out_cache_path(out_path).read_text(encoding='utf-8')

    def check_for_meta_change(self, nav_items: list):
        # If the metadata for any item changes (such as the title), all pages may need to regenerated
        hasher = hashlib.new(self.HASH_ALGORITHM)
        for item in nav_items:
            hasher.update(bytes(dumps(item.as_dict()), 'utf-8'))
        current_hash = hasher.hexdigest()
        if self.hashes['meta'] is None:
            self.hashes['meta'] = current_hash
        elif self.hashes['meta'] != current_hash:
            self.invalidate()
            self.hashes['meta'] = current_hash

    def memo_output(self, rel_path: str, out_path: Path, hashed_content: str):
        self.hashes['items'][rel_path] = hashed_content
        cache_path = self.get_out_cache_path(out_path)
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        cache_path.write_text(out_path.read_text(encoding='utf-8'), encoding='utf-8')

    def dump(self):
        self.hash_file.write_text(dumps(self.hashes), encoding='utf-8')





from rjsmin import jsmin

from menagerie.Items.Meta.MetaItem import MetaItem
from menagerie.Items.MinifiedItemMixin import MinifiedItemMixin


class JSMetaScript(MinifiedItemMixin, MetaItem):

    extensions = ('js',)
    out_extension = 'min.js'
    minify_key = 'js'

    def minify(self, content: str) -> str:
        return jsmin(content)





from pathlib import Path

from menagerie.Items.Managers.AbstractManager import AbstractManager
from menagerie.Items.Meta.JSONMetaItem import JSONMetaItem
from menagerie.Items.Meta.JSMetaItem import JSMetaScript
from menagerie.Items.Meta.CSSMetaItem import CSSMetaItem
from menagerie.Items.Meta.MetaItem import MetaItem
from menagerie.Items.Meta.XMLMetaItem import XMLMetaItem


class MetaManager(AbstractManager):
    item_types = (MetaItem, XMLMetaItem)
    root_dir = ''

    def find(self):
        pass

    def initialize(self):
        self.items = (
            XMLMetaItem(self, Path('sitemap.jinja2'), 'xml'),
            XMLMetaItem(self, Path('browserconfig.jinja2'), 'xml'),
            JSONMetaItem(self, Path('manifest.jinja2'), 'webmanifest'),
            JSMetaScript(self, Path('base.js'), 'min.js'),
            JSMetaScript(self, Path('schema.js'), 'min.js'),
            CSSMetaItem(self, Path('schema.css'), 'min.css'),
            MetaItem(self, Path('robots.jinja2'), 'txt'),
        )

    def generate(self):
        self.base_env.globals['pages'] = self.gen.shared_info['pages']
        self.base_env.filters['static'] = self.gen.shared_info['static_filter']
        for item in self.items:
            item.generate()



import sass

from menagerie.Items.Static.StaticCSS import StaticCSS



class StaticSCSS(StaticCSS):
    extensions = ('scss', 'sass')
    
    def transform(self, content: str) -> str:
        return sass.compile(string=content)




from rcssmin import cssmin

from menagerie.Items.Meta.MetaItem import MetaItem
from menagerie.Items.MinifiedItemMixin import MinifiedItemMixin


class CSSMetaItem(MinifiedItemMixin, MetaItem):

    minify_key = 'css'
    out_extension = 'min.css'
    extensions = ('css',)

    def minify(self, content: str) -> str:
        return cssmin(content)



import re
from pathlib import Path

from markdown import Markdown, Extension
from markdown.extensions.attr_list import AttrListExtension, get_attrs
from markdown.treeprocessors import Treeprocessor
from markdown.extensions.fenced_code import FencedBlockPreprocessor
from markdown.extensions.admonition import AdmonitionProcessor

from menagerie.Items.Pages.AbstractPage import AbstractPage

__all__ = ('MDPage',)

HEADER_BORDER_CLASSES = ' border-bottom border-{{ settings[\'themes\'][\'theme-opposite\'] }} pb-1 mb-1'

ADDITIONAL_CLASSES: dict[str, str] = {
    'img': "img-fluid rounded mx-auto d-flex",
    'table': "table table-striped",
    'h1': 'border-2' + HEADER_BORDER_CLASSES,
    'h2': 'border-1' + HEADER_BORDER_CLASSES
}


class MTreeProcessor(Treeprocessor):

    def __init__(self, page):
        super().__init__()
        self.page = page

    def run(self, node):
        for child in node:
            self.page.process_node(child)
        return node


class MenagerieAdmonitions(AdmonitionProcessor):

    CLASSNAME = "alert"
    CLASSNAME_TITLE = "h4"


def _raise_serialization_error(text):  # pragma: no cover
    raise TypeError(
        "cannot serialize {!r} (type {})".format(text, type(text).__name__)
        )


def _escape_attrib_html(text):
    # escape attribute value
    try:
        if "&" in text:
            # Only replace & when not part of an entity
            text = re.compile(r'&(?!(?:#\d+|#x[\da-f]+|[\da-z]+);)', re.I).sub('&amp;', text)
        if "<" in text:
            text = text.replace("<", "&lt;")
        if ">" in text:
            text = text.replace(">", "&gt;")
        if "\"" in text:
            text = text.replace("\"", "&quot;")
        return text
    except (TypeError, AttributeError):  # pragma: no cover
        _raise_serialization_error(text)


class MenagerieFencedCode(FencedBlockPreprocessor):

    def run(self, lines):
        if not self.checked_for_deps:
            for ext in self.md.registeredExtensions:
                if isinstance(ext, AttrListExtension):
                    self.use_attr_list = True

            self.checked_for_deps = True

        text = "\n".join(lines)
        while 1:
            m = self.FENCED_BLOCK_RE.search(text)
            if m:
                lang, id, classes, config = None, '', [], {}
                if m.group('attrs'):
                    id, classes, config = self.handle_attrs(get_attrs(m.group('attrs')))
                    if len(classes):
                        lang = classes.pop(0)
                else:
                    if m.group('lang'):
                        lang = m.group('lang')

                id_attr = lang_attr = class_attr = kv_pairs = ''
                if lang:
                    prefix = self.config.get('lang_prefix', 'language-')
                    lang_attr = f' class="{prefix}{_escape_attrib_html(lang)}"'
                if classes:
                    class_attr = f' class="{_escape_attrib_html(" ".join(classes))} position-relative"'
                else:
                    class_attr = f' class="position-relative"'
                if id:
                    id_attr = f' id="{_escape_attrib_html(id)}"'
                if self.use_attr_list and config and not config.get('use_pygments', False):
                    # Only assign key/value pairs to code element if attr_list ext is enabled, key/value pairs
                    # were defined on the code block, and the `use_pygments` key was not set to True. The
                    # `use_pygments` key could be either set to False or not defined. It is omitted from output.
                    kv_pairs = ''.join(
                        f' {k}="{_escape_attrib_html(v)}"' for k, v in config.items() if k != 'use_pygments'
                    )
                code = self._escape(m.group('code'))
                code = f'<pre{id_attr}{class_attr}><button class="menagerie-copy-button btn btn-sm btn-outline-primary position-absolute" style="top: 5px; right: 5px;" aria-label="Copy"><i class="bi bi-clipboard2"></i></button><code{lang_attr}{kv_pairs}>{code}</code></pre>'

                placeholder = self.md.htmlStash.store(code)
                text = f'{text[:m.start()]}\n{placeholder}\n{text[m.end():]}'
            else:
                break
        return text.split("\n")


class MenagerieMarkdownExtension(Extension):

    def __init__(self, page, **kwargs):
        self.page = page
        super().__init__(**kwargs)

    def extendMarkdown(self, md: Markdown) -> None:
        md.registerExtension(self)
        self.tree = MTreeProcessor(self.page)
        self.tree.md = md
        self.tree.config = self.getConfigs()
        md.treeprocessors.add('menagerie_tree', self.tree, '_end')
        md.parser.blockprocessors.register(MenagerieAdmonitions(md.parser), 'm-admonition', 105)
        md.preprocessors.register(MenagerieFencedCode(md, {}), 'm-copy-button', 26)


class MDPage(AbstractPage):
    extensions = ('md', 'markdown')

    md: Markdown

    def process_node(self, node) -> None:
        if node.tag in ADDITIONAL_CLASSES.keys():
            node.set('class', ADDITIONAL_CLASSES[node.tag])
        if node.tag == 'img':
            possible_match = re.match(r"{{ ?[\"'](.*?)['\"]\|static.*?}}", node.get('src'))
            if possible_match:
                image_path = possible_match.group(1)
                size = self.manager.gen.shared_info['image_sizes'].get(str(Path(image_path).as_posix()), (100, 100))
                node.set('width', str(size[0]))
                node.set('height', str(size[1]))
        elif node.tag == 'a':
            url = node.get('href')
            if url.startswith('http') and (url.startswith(self.manager.gen.settings['base_url']) is False):
                node.set('target', "_blank")
                node.set('rel', "noopener")
        for child in node:
            self.process_node(child)

    def __init__(self, manager, path: Path):
        super(MDPage, self).__init__(manager, path)
        md_settings = {
            'extensions': ['extra', 'toc', 'meta', MenagerieMarkdownExtension(self)],
            'output-format': 'html5'
        }
        self.md = Markdown(**md_settings)

    def load_metadata(self) -> None:
        self.md.convert(self.get_content()).replace('&quot;', '"')
        self.meta.update({k: None if v is None else v[0] for k, v in self.md.Meta.items()})
        self.meta['table_of_contents'] = self.md.toc_tokens

    def inner_render(self, content: str) -> str:
        converted = self.md.convert(self.get_content()).replace('&quot;', '"')
        return self.manager.env.from_string(converted).render(page=self, **self.manager.context)



import argparse
from pathlib import Path

from menagerie.generate import generate
from menagerie.new_project import new_project

__version__ = '0.1.13'


def get_parser():
    # Root Parser
    parser = argparse.ArgumentParser(description="Run a menagerie command")
    parser.add_argument("-v", "--version", action='version', version=__version__)
    subparsers = parser.add_subparsers(title='command', dest='command', required=True)

    # New Project Parser
    new_proj_parser = subparsers.add_parser('new-project', help="Start a new menagerie site")
    new_proj_parser.add_argument('name', help="The new folder to create and place the new project in",
                                 default="NewProject")

    # Generate Parser
    gen_parser = subparsers.add_parser('generate', help="Generate an existing menagerie site")
    gen_parser.add_argument("-c", "--config", help="Specify a different config file to use", default="config.json",
                            type=Path)
    gen_parser.add_argument("--clear-cache", action="store_true", help="Clear the cache before generating")

    return parser


def execute_from_commandline(args: list[str] = None):
    parser = get_parser()
    if args is None:
        parsed_args = parser.parse_args()
    else:
        parsed_args = parser.parse_args(args)
    if parsed_args.command == 'generate':
        generate(Path(parsed_args.config), parsed_args.clear_cache)
    elif parsed_args.command == 'new-project':
        new_project(parsed_args.name)



import re


def remove_preceding_slash(path: str):
    return path[1:] if len(path) > 1 and path[0] == "/" else path


def camel_to_pretty(raw):
    return ' '.join(re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', raw))


def pretty_title(raw: str) -> str:
    if '_' in raw:
        return ' '.join(x[0].upper() + x[1:] for x in raw.split('_'))
    elif any(x.isupper() for x in raw):
        if raw[0].islower():
            new_raw = raw[0].upper() + raw[1:]
        else:
            new_raw = raw
        return camel_to_pretty(new_raw)
    else:
        return raw[0].upper() + raw[1:]
